/*
SimpleAlgebra.cpp
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "SimpleAlgebra.h"

namespace algebra
{
void mat::random_unitary(const std::vector <double>& phi_v, const std::vector <double>& psi_v, const std::vector <double>& chi_v)
{
	int dim = getRows();
	algebra::mat E, temp, temp2;
	temp.create(dim);
	int index = 0;
	Complex z;
	ident();
	for(int j=1; j<=dim-1; ++j)
	{
		E.create(dim);
		E.ident();
		for(int i=j; i>=1; --i)
		{
			int I = i;
			int J = j+1;
			double psi, phi, chi;
			psi = psi_v[index];
			phi = phi_v[index];
			if(I == 1)
				chi = chi_v[i-1];
			else
				chi = 0;

			++index;
			temp.ident();
			z = algebra::ComplexHelper::construct(psi);
			z.real *= cos(phi);
			z.imag *= cos(phi);
			temp(I-1, I-1) = z;

			z = algebra::ComplexHelper::construct(chi);
			z.real *= sin(phi);
			z.imag *= sin(phi);
			temp(I-1, J-1) = z;

			z = algebra::ComplexHelper::construct(-chi);
			z.real *= -sin(phi);
			z.imag *= -sin(phi);
			temp(J-1, I-1) = z;

			z = algebra::ComplexHelper::construct(-psi);
			z.real *= cos(phi);
			z.imag *= cos(phi);
			temp(J-1, J-1) = z;

			E.multiply(temp, temp2);
			E = temp2;
		}

		multiply(E, temp2);
		*this = temp2;
	}
}

void mat::tensor(mat* B, mat* output)
{
	if(!B || !output)
		return;

	output->create(getRows()*B->getRows(), getCols()*B->getCols());

	for(int r=0; r<output->getRows(); ++r)
	{
		for(int c=0; c<output->getCols(); ++c)
		{
			int aRow = r / B->getRows();
			int aCol = c / B->getCols();
			int bRow = r % B->getRows();
			int bCol = c % B->getCols();

			output->setDirect(r, c, ComplexHelper::mult(getDirect(aRow, aCol), B->getDirect(bRow, bCol)));
		}
	}
}

bool mat::multiply(mat& B, mat& output)
{
	if(getCols() != B.getRows())
		return false;

	output.create(getRows(), B.getCols());
	Complex16 sum, prod;
	for(int row=0; row<output.getRows(); ++row)
	{
		for(int col=0; col<output.getCols(); ++col)
		{
			sum.real = 0;
			sum.imag = 0;
			for(int k=0; k<getCols(); ++k)
			{
				prod = ComplexHelper::mult((*this)(row,k), B(k,col));
				sum = ComplexHelper::add(sum, prod);
			}

			output(row, col) = sum;
		}
	}
	return true;
}

void mat::dot(mat* B, Complex16& output)
{
	if(!B)
		throw std::string("dot product: B is NULL");
	if(B->getSize() != getSize())
		throw std::string("Dot product error 1\n");
	if(getRows() != 1 && getCols() != 1)
		throw std::string("Dot product error 2\n");
	if(B->getCols() != 1 && B->getCols() != 1)
		throw std::string("Dot product error 3\n");

	Complex16 sum, prod;
	sum.imag = 0;
	sum.real = 0;
	for(int i=0; i<n; ++i)
	{
		prod = ComplexHelper::mult(elements[i], B->getElements()[i]);
		sum = ComplexHelper::add(sum, prod);
	}

	output = sum;
}
}